/*
 * Bresenham Line Drawing Routine
 * Erik Nygren (nygren@mit.edu)
 * 1994.09.15
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "image.h"

#define PLOT(bm,x,y) (((bm)->data)[(x)+(y)*(bm)->xres] = (color))

/*
 * image_drawline -- 
 *    Implementation of Bresenham's algorithm.
 *    Draws lines from xFrom,yFrom to xTo,yTo
 */
void
imageplane_drawline(IPlane *bm, uchar color, 
		    int xFrom, int yFrom, int xTo, int yTo)
{
  int dx, dy;                 /* The line slope */
  register int x, y;          /* X and Y positions */
  register int d;             /* The decision value */
  register int incrA, incrB;  /* Incriment to d used for choices A and B */
  double M, B;                /* Slope and intercept for clipping */
  int vert;                   /* Line is vertical */
  int clipx, clipy;

  if (!bm) 
    { fprintf(stderr, "imageplane_drawline: NULL bm\n"); return; }
  clipx = bm->xres-1;
  clipy = bm->yres-1;

  dx = xTo - xFrom;
  dy = yTo - yFrom;
  if (dx == 0) vert = 1;
  else 
    {
      vert = 0;
      M = (double)dy/(double)dx;
      B = (double)yFrom - M*(double)xFrom;
    }

  /* Do clipping on (0,0,clipx,clipy) */
  if (vert)
    {
      if (xFrom < 0 || xFrom > clipx)
	return;
      if (yFrom < 0)
	yFrom = 0;
      if (yFrom > clipy)
	yFrom = clipy;
      if (yTo < 0)
	yTo = 0;
      if (yTo > clipy)
	yTo = clipy;
    }
  else
    {
      if ((xFrom < 0 && xTo < 0)
	  || (xFrom > clipx && xTo > clipx) 
	  || (yFrom < 0 && yTo < 0)
	  || (yFrom > clipy && yTo > clipy))
	{
	  return;
	}
      if (xFrom < 0)
	{ xFrom = 0; yFrom = B; }
      else if (xTo < 0)
	{ xTo = 0; yTo = B; }
      if (xFrom > clipx)
	{ xFrom = clipx; yFrom = M*clipx+B; }
      else if (xTo > clipx)
	{ xTo = clipx; yTo = M*clipx+B; }
      if (yFrom < 0)
	{ yFrom = 0; xFrom = -B/M; }
      else if (yTo < 0)
	{ yTo = 0; xTo = -B/M; }
      if (yFrom > clipy)
	{ yFrom = clipy; xFrom = (clipy-B)/M; }
      else if (yTo > clipy)
	{ yTo = clipy; xTo = (clipy-B)/M; }
    }

  /* Now draw the line */
  x = xFrom; y = yFrom;

  PLOT(bm,x,y);   /* Draw the first pixel */

  if (abs(dy) <= abs(dx))
    {
      if (dx > 0)
	{
	  if (dy > 0)
	    { /* Generally NE */
	      incrA = 2*dy;
	      incrB = 2*(dy-dx);
	      d = 2*dy - dx;      
	      while (x < xTo)
		{ 
		  if (d <= 0) { d += incrA; x++; }
		  else        { d += incrB; x++; y++; }
		  PLOT(bm,x,y);
		}
	    }
	  else /* dy < 0 */
	    {  /* Generally SE */
	      incrA = 2*dy;
	      incrB = 2*(dy+dx);
	      d = 2*dy + dx;      
	      while (x < xTo) 
		{ 
		  if (d >= 0) { d += incrA; x++; }
		  else        { d += incrB; x++; y--; }
		  PLOT(bm,x,y);
		}
	    }
	}
      else /* dx < 0 */
	{
	  if (dy > 0)
	    { /* Generally NW */
	      incrA = -2*dy;
	      incrB = 2*(-dy-dx);
	      d = -2*dy - dx;      
	      while (x > xTo)
		{ 
		  if (d > 0) { d += incrA; x--; }
		  else       { d += incrB; x--; y++; }
		  PLOT(bm,x,y);
		}
	    }
	  else /* dy < 0 */
	    {  /* Generally SW */
	      incrA = -2*dy;
	      incrB = 2*(-dy+dx);
	      d = -2*dy + dx;    
	      while (x > xTo)
		{ 
		  if (d < 0) { d += incrA; x--; }
		  else       { d += incrB; x--; y--; }
		  PLOT(bm,x,y);
		}
	    }
	}
    }
  else   /* abs(dy) > abs(dx) */
    {
      if (dy > 0)
	{
	  if (dx > 0)
	    { /* Generally NE */
	      incrA = 2*dx;
	      incrB = 2*(dx-dy);
	      d = 2*dx - dy;      
	      while (y < yTo)
		{ 
		  if (d <= 0) { d += incrA; y++; }
		  else        { d += incrB; y++; x++; }
		  PLOT(bm,x,y);
		}
	    }
	  else /* dx < 0 */
	    {  /* Generally NW */
	      incrA = 2*dx;
	      incrB = 2*(dx+dy);
	      d = 2*dx + dy;      
	      while (y < yTo) 
		{ 
		  if (d >= 0) { d += incrA; y++; }
		  else        { d += incrB; y++; x--; }
		  PLOT(bm,x,y);
		}
	    }
	}
      else /* dy < 0 */
	{
	  if (dx > 0)
	    { /* Generally SE */
	      incrA = -2*dx;
	      incrB = 2*(-dx-dy);
	      d = -2*dx - dy;      
	      while (y > yTo)
		{ 
		  if (d > 0) { d += incrA; y--; }
		  else       { d += incrB; y--; x++; }
		  PLOT(bm,x,y);
		}
	    }
	  else /* dx < 0 */
	    {  /* Generally SW */
	      incrA = -2*dx;
	      incrB = 2*(-dx+dy);
	      d = -2*dx + dy;    
	      while (y > yTo)
		{ 
		  if (d < 0) { d += incrA; y--; }
		  else       { d += incrB; y--; x--; }
		  PLOT(bm,x,y);
		}
	    }
	}
    }
}
