/* $Header: /home/nygren/courses/6.866/project/src/RCS/image.c,v 1.8 1996/03/02 17:09:59 nygren Exp $ */
/*
 * FILE:    image.c
 * AUTHOR:  Erik Nygren.
 * PURPOSE: Routines for loading and manipulating images.
 *          Written for 6.866 term project, Fall 1995.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "image.h"
#include "util.h"

#undef DEBUG

#define BUFLEN 3072

/* image_create --
 *    Creates an image with size xres*yres and np planes.
 *    Returns the Image or NULL on failure.
 */
Image *image_create(int xres, int yres, int np)
{
  Image  *im;
  int     planectr;  /* plane counter */
  IPlane *ip;        /* current image plane */

  im = (Image*)xcalloc("image_create", 1, sizeof(Image));
  im->xres = xres;
  im->yres = yres;
  im->xorigin = xres/2;
  im->yorigin = yres/2;
  im->source = "image_create";
  im->np = np;
  if (np<=0 || np>MAX_IMAGEPLANES)
    { fprintf(stderr, "image_create: nplanes out of bounds\n");
      exit(-1); }
  for (planectr=0; planectr<np; planectr++)
    {
      ip = (IPlane*)xcalloc("image_create", 1, sizeof(IPlane));
      im->p[planectr] = ip;
      ip->xres = xres;
      ip->yres = yres;
      ip->xorigin = xres/2;
      ip->yorigin = yres/2;
      ip->data = (uchar*)xcalloc("image_create", xres*yres, sizeof(char));
    }
  return(im);
}

/* image_destroy ---
 *    Frees the data associated with an image.
 */
void 
image_destroy(Image *im)
{
  int ctr;
  
  if (!im) return;
  for (ctr=0; ctr < im->np; ctr++)
    {
      xfree(im->p[ctr]->data);
      xfree(im->p[ctr]);
    }
  xfree(im);
}

/* image_loadpnm ---
 *     Loads a PNM image from fd.  (filename is for errors).
 *     Returns the new Image or NULL if an error 
 */
Image *
image_loadpnm(int fd, char *filename)
{
  char type;         /* pnm type */
  char buf[BUFLEN]; 
  int xres, yres, maxval, np, asciibody;
  Scanner *scan;
  Image *im;

  scan = scanner_create(fd);
  
  /* Get magic */
  scanner_gettoken(scan, buf, BUFLEN);
  if (scanner_eof(scan))
    { fprintf(stderr, "image_loadpnm: premature end of file in %s\n", 
	      filename);
      return(NULL); }
  if (buf[0] != 'P' || buf[2])
    { fprintf(stderr, "image_loadpnm: %s has bad magic number\n", 
	      filename);
      return(NULL); }
  type = buf[1];

  if (type=='2')  /* ASCII PGM */
    { np=1; asciibody=1; }
  else if (type=='5') /* RAW PGM */
    { np=1; asciibody=0; }
  else if (type=='3') /* ASCII PPM */
    { np=3; asciibody=1; }
  else if (type=='6') /* RAW PPM */
    { np=3; asciibody=0; }
  else 
    { fprintf(stderr, "image_loadpnm: %s is not in a supported format\n", 
	      filename);
      return(NULL); }

  scanner_gettoken(scan, buf, BUFLEN);
  if (scanner_eof(scan))
    { fprintf(stderr, "image_loadpnm: premature end of file in %s\n", 
	      filename);
      return(NULL); }
  xres = atoi(buf);
  if (xres>=(BUFLEN/np)-1)
    { fprintf(stderr, "image_loadpnm: max image size %d while loading %s\n", 
	      BUFLEN/np-1, filename);
      return(NULL); }

  scanner_gettoken(scan, buf, BUFLEN);
  if (scanner_eof(scan))
    { fprintf(stderr, "image_loadpnm: premature end of file in %s\n", 
	      filename);
      return(NULL); }
  yres = atoi(buf);

  scanner_gettoken(scan, buf, BUFLEN);
  if (scanner_eof(scan))
    { fprintf(stderr, "image_loadpnm: premature end of file in %s\n", 
	      filename);
      return(NULL); }
  maxval = atoi(buf);
  if (maxval!=255)  
    { fprintf(stderr, "image_loadpnm: maxval of %d in %s is not supported\n", 
	      maxval, filename);
      return(NULL); }

  if (!(im = image_create(xres, yres, np)))
    { fprintf(stderr, "image_loadpnm: failed to create image for %s\n",
	      filename);
      return(NULL); }
  im->source = string_dup(filename);

  if (asciibody)
    {
      int i,p;

      for (i=0; i<xres*yres; i++)
	for (p=0; p<np; p++)
	  {
	    scanner_gettoken(scan, buf, BUFLEN);
	    if (scanner_eof(scan))
	      { fprintf(stderr, 
			"image_loadpnm: premature end of file in %s\n", 
			filename);
		return(NULL); }
	    im->p[p]->data[i] = (uchar)atoi(buf);
	  }
    }
  else /* raw body */
    {
      int pc, xpos, ypos;
      uchar *pptr[MAX_IMAGEPLANES], *tp;
      
      for (pc=0; pc<np; pc++)
	  pptr[pc] = im->p[pc]->data;
      for (ypos=0; ypos<yres; ypos++)
	{
	  if (!read(fd, buf, xres*np))
	    { fprintf(stderr, 
		      "image_loadpnm: premature end of file in %s\n", 
		      filename);
	      return(NULL); }
	  tp = (uchar*)buf;
	  for (xpos=0; xpos<xres; xpos++)
	    for (pc=0; pc<np; pc++)
	      *((pptr[pc])++) = *(tp++);
	}
    }

  fprintf(stderr, "image_loadpnm: loaded %dx%dx%d image %s\n", 
	  xres, yres, np, filename);
  return(im);
}

Image *image_copy(Image *im)
{
  return(im);
}

/* image_writepnm ---
 *    Writes image im to fd in RAW format.  Returns 0 on failure.
 */
int image_writepnm(Image *im, int fd)
{
  char  sbuf[128];

  if (im->np == 1)
    write(fd, "P5", 2);
  else if (im->np == 3)
    write(fd, "P6", 2);
  else 	
    { fprintf(stderr, "image_writepnm: number of planes unsupported\n");
      return(0); }
  
  sprintf(sbuf, "\n# Image generated from source %s\n", im->source);
  write(fd, sbuf, strlen(sbuf));
  sprintf(sbuf, "%d %d 255\n", im->xres, im->yres);
  write(fd, sbuf, strlen(sbuf));

  if (im->np==1)
    {
      uchar *pos = im->p[0]->data;
      int pixleft = im->xres*im->yres;   /* pixels remaining */

      while (pixleft > 0)
	{
	  write(fd, pos, MIN(BUFLEN,pixleft));
	  pos += BUFLEN;
	  pixleft -= BUFLEN;
	}
    }
  else /* np==3 */
    {
      uchar *outbuf, *obpos;
      uchar *pos0 = im->p[0]->data;
      uchar *pos1 = im->p[1]->data;
      uchar *pos2 = im->p[2]->data;
      int x,y;

      outbuf = (uchar*)xcalloc("image_writepnm", im->xres, 3*sizeof(uchar));
      for (y=0; y < im->yres; y++)
	{
	  obpos = outbuf;
	  for (x=0; x < im->xres; x++)
	    {
	      *(obpos++) = *(pos0++);
	      *(obpos++) = *(pos1++);
	      *(obpos++) = *(pos2++);
	    }
	  write(fd, outbuf, im->xres*3);
	}
      xfree(outbuf);
    }
  return(1);
}

/* Gets the value of a pixel at a location relative to the
 * upper left corner of the image. */
uchar
iplane_getval(IPlane *ip, int x, int y)
{
  if (x >= 0  &&  y >= 0  &&  x < ip->xres  &&  y < ip->yres)
    return((ip->data)[x+y*ip->xres]);
  else
    {
      fprintf(stderr, "iplane_getval: %d,%d out of bounds\n",x,y);
      return(0);
    }
}

/* Gets the value of a pixel at a location relative to the
 * origin of the image. 
 */
uchar
iplane_getval_rel(IPlane *ip, int x, int y)
{
  x += ip->xorigin;
  y += ip->yorigin;
  if (x >= 0  &&  y >= 0  &&  x < ip->xres  &&  y < ip->yres)
    return((ip->data)[x+y*ip->xres]);
  else
    return(0);
}

/* Gets the packed value of a pixel at a location relative to the
 * origin of the image.  Returns -1 on out of bounds.
 */
ulong32
image_getval_rel(Image *im, int x, int y)
{
  int i, off;
  ulong32 result = 0;

  x += im->xorigin;
  y += im->yorigin;
  if (x >= 0  &&  y >= 0  &&  x < im->xres  &&  y < im->yres)
    off = x+y*im->xres;
  else
    return(-1);

  for (i = 0; i < im->np; i++)
    result = result | ((ulong32)((im->p[i]->data)[off]))<<(i*8);
  return(result);
}

/* Sets the packed value of a pixel at a location relative to the
 * origin of the image.  Returns -1 on out of bounds.
 */
void
image_setval_rel(Image *im, int x, int y, ulong32 val)
{
  int i, off;

  x += im->xorigin;
  y += im->yorigin;
  if (x >= 0  &&  y >= 0  &&  x < im->xres  &&  y < im->yres)
    off = x+y*im->xres;
  else
    return;

  for (i = 0; i < im->np; i++)
    {
      ((im->p[i]->data)[off]) = (uchar)(0xff&val);
      val >>= 8;
    }
}



/**************** FILE SCANNER UTILITIES **************/

Scanner *
scanner_create(int fd)
{
  Scanner *s;

  s = xcalloc("scanner_create", 1, sizeof(Scanner));
  s->fd = fd;
  s->eof = !read(s->fd, &(s->cur), 1);
  return(s);
}

/* Gets the next token, eating any leading whitespace */
void
scanner_gettoken(Scanner *s, char *buf, int bufsize)
{
  int ctr=0;

  scanner_eatwhitespace(s);
  while (!(s->eof) && !isspace(s->cur) && (s->cur != '#') && (ctr<bufsize))
    {
      buf[ctr++] = s->cur;;
      s->eof = !read(s->fd, &(s->cur), 1);
    }
  buf[ctr] = '\0';
#ifdef DEBUG
  printf("got token <%s>\n",buf);
#endif
}

/* Eats up whitespace and returns when done or eof. */
void
scanner_eatwhitespace(Scanner *s)
{
  int state = 0;

  while (!(s->eof) && (state != -1))
    {
      switch (state)
	{
	case 0:  /* in whitespace */
	  if (s->cur == '#')
	    {
	      state = 1;  /* goto comment */
	      s->eof = !read(s->fd, &(s->cur), 1);      
	    }
	  else if (!isspace(s->cur))
	    state = -1;
	  else
	    s->eof = !read(s->fd, &(s->cur), 1);      
	  break;

	case 1:  /* in comment */
	  if (s->cur == '\n')
	    state = 0;  /* goto whitespace */
	  s->eof = !read(s->fd, &(s->cur), 1);
	  break;
	}
    }
}


