/* 
 * FILE:    xdisplay.c
 * PURPOSE: display a greyscale image in an X window
 * AUTHOR:  Erik Nygren
 */

#include <stdio.h>               
#include <stdlib.h>               
#include <X11/Xlib.h>            /* Xlib include files */
#include <X11/Xutil.h>           /* Must have these included */
#include <X11/Xatom.h>
#include "image.h"
#ifdef OPFLOW_SUPPORT
#include "opflow.h"
#endif
#include "util.h"
#include "imagedisplayer.h"

static Display *id_xdisplay = NULL;

/* Initializes stuff, like opening display.  Returns 0 on failure. */
int
imagedisplayer_init(void)
{
  char *display_name = ":0", *display_env;
  
  if (NULL != (display_env = getenv("DISPLAY")))
    display_name = display_env;
  
  if ((id_xdisplay=XOpenDisplay(display_name)) == NULL) 
    {
      fprintf(stderr, "imagedisplayer_init: cannot connect to X server");
      if (display_env == NULL)
	fprintf(stderr, ", 'DISPLAY' environment variable not set.\n");
      else
	fprintf( stderr, " %s\n", XDisplayName(display_name));
      return(0);
    }
  return(1);
}

/* Creates an X window to display a greyscale image "im" in.
 * Returns NULL if fails. */
ImageDisplayer *
imagedisplayer_create(int xres, int yres, char *window_name)
{
  XSizeHints size_hints;
  int screen;
  Window rootwin;
  unsigned long black_color, white_color;
  static Colormap colmap = 0;
  int depth;
  XSetWindowAttributes xswa;
  unsigned int k;
  XColor colors[256];
  XVisualInfo visinfo;
  Atom ATOM_WM_COLMAP_WIN;
  ImageDisplayer *id;
  Cursor wincursor;
  Font wincursorfont;
  XGCValues xgcv;

  if (!(id = (ImageDisplayer*)malloc(sizeof(ImageDisplayer))))
    {
      fprintf(stderr, "Unable to allocate ImageDisplayer\n");
      return(NULL);
    }
  id->xres = xres;
  id->yres = yres;
  if (!(id->data = malloc(id->xres*id->yres)))
    {
      fprintf(stderr, "Unable to allocate data for ImageDisplayer\n");
      return(NULL);
    }

  screen = DefaultScreen(id_xdisplay);
  rootwin = RootWindow(id_xdisplay, screen);

  /**** SET UP COLORS ****/
  black_color = BlackPixel(id_xdisplay, screen);
  white_color = WhitePixel(id_xdisplay, screen);

  depth = DefaultDepth(id_xdisplay,screen); 
  /* Get a 256 color PseudoColor visual */
  if (XMatchVisualInfo(id_xdisplay,screen,8,PseudoColor,&visinfo))
    {
      colmap = XCreateColormap(id_xdisplay, rootwin,
			       visinfo.visual, 
			       AllocAll);

      /* Make 0x000 be red */
      colors[0].pixel = 0;
      colors[0].red = 65000;
      colors[0].green = colors[0].blue = 0;
      colors[0].flags = DoRed | DoGreen | DoBlue;

      /* Set up colormap for greyscale */
      for(k=1; k<256; k++)
	{
	  colors[k].pixel = k;
	  colors[k].red = colors[k].green = colors[k].blue 
	    = (u_short)(k*65000/256);
	  colors[k].flags = DoRed | DoGreen | DoBlue;
	}
      XStoreColors(id_xdisplay,colmap,colors,256);

      /* Get window attributes */
      xswa.colormap = colmap;
    }	
  else
    {
      fprintf(stderr,
	  "Sorry, apropriate visuals are not available on this display...\n");
      return(NULL);
    }

  id->win = XCreateWindow(id_xdisplay, rootwin,
			  0, 0, xres, yres, 0, depth, InputOutput,
			  visinfo.visual, (colmap ? CWColormap : 0), &xswa);
  XSetWindowBackground(id_xdisplay, id->win, white_color);

  /* Set colormap for this window and for parent window */
  XSetWindowColormap(id_xdisplay, id->win, colmap);

  /* Set properties for window manager before mapping it */
  XSetStandardProperties(id_xdisplay, id->win, window_name, window_name, None,
			 0, 0, &size_hints);
  XSelectInput(id_xdisplay, id->win, ExposureMask | KeyPressMask |
	       ButtonPressMask | StructureNotifyMask);

  /* Tell windowmanager that window has its own colormap */
  ATOM_WM_COLMAP_WIN = XInternAtom(id_xdisplay,"WM_COLMAP_WINDOWS", False);
  XChangeProperty(id_xdisplay,rootwin,ATOM_WM_COLMAP_WIN,XA_WINDOW,32,
		  PropModeReplace,
		  (unsigned char *)&rootwin, 1);
  XSync(id_xdisplay, True);

  wincursorfont = XLoadFont(id_xdisplay, "cursor");
  wincursor = XCreateGlyphCursor(id_xdisplay, wincursorfont, wincursorfont,
				 0x22, 0x22, &(colors[0]), &(colors[1]));
  XDefineCursor(id_xdisplay, id->win, wincursor);
  

  /* Put window up */
  XMapWindow(id_xdisplay, id->win);
  XFlush(id_xdisplay);

  xgcv.foreground = 0;
  xgcv.background = 1;
  id->gc = XCreateGC (id_xdisplay, id->win, GCForeground | GCBackground,
		      &xgcv);

  id->ximage = XCreateImage(id_xdisplay, visinfo.visual, 8, ZPixmap, 0, NULL,
			    xres, yres, 8, 0);
  
  xgcv.foreground = 0;
  xgcv.background = 1;
  id->pm = XCreatePixmap(id_xdisplay, id->win, xres, yres, 8);
  id->pmgc = XCreateGC (id_xdisplay, id->pm, GCForeground | GCBackground,
			&xgcv);

  id->ximage->data = (char*)id->data;
  return(id);
}

void
imagedisplayer_destroy(ImageDisplayer *id)
{
  if (!id) return;
  XDestroyWindow(id_xdisplay, id->win);	/* Also unmaps the window */
  XDestroyImage(id->ximage);            /* Also frees id->data */
  XFreeGC(id_xdisplay, id->pmgc);
  XFreeGC(id_xdisplay, id->gc);
  XFreePixmap(id_xdisplay, id->pm);
  XFlush(id_xdisplay);
}

void
imagedisplayer_putimage(ImageDisplayer *id, Image *im)
{
  int i;

  if ((id->xres != im->xres) || (id->yres != im->yres))
    { fprintf(stderr, "imagedisplayer_putimage: resolutions do not match\n");
      return; }
  if (im->np != 1)
    { fprintf(stderr, "imagedisplayer_putimage: too many planes\n");
      return; }
  memcpy(id->data, im->p[0]->data, id->xres*id->yres);

  /* Convert 0x00 to 0x01 */
  for (i=0; i < id->xres*id->yres; i++)
    if ((id->data)[i] == 0x00) 
      (id->data)[i] = 0x01;

  XPutImage(id_xdisplay, id->pm, id->pmgc, id->ximage, 0, 0, 0, 0, 
            im->xres, im->yres);
  XCopyArea(id_xdisplay, id->pm, id->win, id->gc, 0, 0, 
            im->xres, im->yres, 0, 0);

  XFlush(id_xdisplay);
}

/* Refreshes window with whatever's in id->pm */
void
imagedisplayer_refresh(ImageDisplayer *id)
{
  XEvent event;                 /* To store event in */

  XCopyArea(id_xdisplay, id->pm, id->win, id->gc, 0, 0, 
            id->xres, id->yres, 0, 0);
  /* Catch up on any more expose events before returning. */
  while (XCheckWindowEvent(id_xdisplay, id->win, ExposureMask, &event));
}

#ifdef OPFLOW_SUPPORT

/* Overlays an optical flow field on top of an image to be displayed */
void
imagedisplayer_putiplane_withflow(ImageDisplayer *id, IPlane *ip,
				  OpFlow *of)
{
  IPlane tip;  /* temp image plane */
  int x, y, xr, yr, u, v, i;

  fprintf(stderr, 
	  "imagedisplayer_putiplane_withflow: currently unsupported\n");
  if ((id->xres != ip->xres) || (id->yres != ip->yres))
    { fprintf(stderr, "imagedisplayer_putimage: resolutions do not match\n");
      return; }
  memcpy(id->data, ip->data, id->xres*id->yres);

  /* Convert 0x00 to 0x01 */
  for (i=0; i < id->xres*id->yres; i++)
    if ((id->data)[i] == 0x00) 
      (id->data)[i] = 0x01;

  tip.data = id->data;
  tip.xres = id->xres;
  tip.yres = id->yres;

  xr = MIN(id->xres, of->xres);
  yr = MIN(id->yres, of->yres);

  /* Draw the optical flow into the image plane */
  for (y = 0; y < yr; y += 5)
    for (x = 0; x < xr; x += 5)
      {
	u = (int)((of->u)[x+y*of->xres]);
	v = (int)((of->v)[x+y*of->xres]);
	imageplane_drawline(&tip, 0, x, y, x+u, y+v);
	imageplane_drawline(&tip, 128, x, y, x, y);
      }

  /* Now render the image */
  XPutImage(id_xdisplay, id->win, id->gc, id->ximage, 0, 0, 0, 0, 
	    ip->xres, ip->yres);
  XFlush(id_xdisplay);
}

#endif

/* Blocks until a button is pressed and stores the location relative
 * to the image displayer in xpos and ypos.  */
void
imagedisplayer_get_buttonpress(ImageDisplayer *id, 
			       int *xpos, int *ypos)
{
  XEvent       event;		/* To store event into */
  XButtonEvent *buttonevent;	/* For accessing coords */

  if (!xpos || !ypos || !id)
    { fprintf(stderr, "imagedisplayer_get_buttonpress: invalid args\n");
      return; }

  while (1)
    {
      XWindowEvent(id_xdisplay, id->win,
		   ExposureMask | ButtonPressMask, &event);
      switch (event.type)
	{
	case Expose:
	  imagedisplayer_refresh(id);
	  break;
	case GraphicsExpose:
	  imagedisplayer_refresh(id);
	  break;
	case ButtonPress:
	  buttonevent = (XButtonEvent*)&event;
	  *xpos = buttonevent->x;
	  *ypos = buttonevent->y;
	  /* Put a red dot at the site of the click. */
	  XDrawPoint(id_xdisplay, id->win, id->gc,  *xpos,   *ypos);
	  XDrawPoint(id_xdisplay, id->pm, id->pmgc, *xpos,   *ypos);
	  XDrawPoint(id_xdisplay, id->win, id->gc,  *xpos+1, *ypos);
	  XDrawPoint(id_xdisplay, id->pm, id->pmgc, *xpos+1, *ypos);
	  XDrawPoint(id_xdisplay, id->win, id->gc,  *xpos-1, *ypos);
	  XDrawPoint(id_xdisplay, id->pm, id->pmgc, *xpos-1, *ypos);
	  XDrawPoint(id_xdisplay, id->win, id->gc,  *xpos,   *ypos+1);
	  XDrawPoint(id_xdisplay, id->pm, id->pmgc, *xpos,   *ypos+1);
	  XDrawPoint(id_xdisplay, id->win, id->gc,  *xpos,   *ypos-1);
	  XDrawPoint(id_xdisplay, id->pm, id->pmgc, *xpos,   *ypos-1);

	  XFlush(id_xdisplay);
	  return;
	}
    }
}
