/* $Header: /home/nygren/courses/6.866/project/src/RCS/linalg.c,v 1.3 1996/03/03 02:07:22 nygren Exp $ */
/*
 * FILE:    linalg.c
 * AUTHOR:  Erik Nygren.
 * PURPOSE: Linear algebra routines.
 *          Written for 6.866 term project, Fall 1995.
 */

#include <stdio.h>
#include <math.h>
#include "linalg.h"
#include "util.h"

#define ZERO_TOLERANCE 1.0e-8
#define ISZERO(x) (fabs(x) < ZERO_TOLERANCE)

/* matrix_create ---
 *    Creates an nrows x ncols matrix.
 */
Matrix *
matrix_create(int nrows, int ncols)
{
  Matrix *mat;

  mat = (Matrix*)xcalloc("matrix_create", 1, sizeof(Matrix));
  mat->el = (double*)xcalloc("matrix_create", nrows*ncols,
			     sizeof(double));
  mat->rows = nrows;
  mat->cols = ncols;
  return(mat);
}

/* matrix_destroy ---
 *    Frees the space used by a matrix.
 */
void
matrix_destroy(Matrix *mat)
{
  if (mat && mat->el)
    {
      xfree(mat->el);
      xfree(mat);
    }
}

/* matrix_create_fromarray ---
 *    Creates an nrows x ncols matrix from a 1D array of data
 *    in row major order (mainly for debugging).
 */
Matrix *
matrix_create_fromarray(int nrows, int ncols, double *data)
{
  Matrix *mat;
  int i;
  
  mat = matrix_create(nrows, ncols);
  for (i = 0; i < nrows*ncols; i++)
    mat->el[i] = data[i];
  return(mat);
}

/* matrix_zero ---
 *    Zeros a matrix A.
 */
void
matrix_zero(Matrix *A)
{
  int i;
  
  for (i = 0; i < A->rows*A->cols; i++)
    (A->el)[i] = 0.0;
}

/* matrix_copy ---
 *    Copies a matrix SRC into DEST
 */
void
matrix_copy(Matrix *DEST, const Matrix *SRC)
{
  int i;

  if (!DEST || !SRC || DEST->rows!=SRC->rows || DEST->cols!=SRC->cols)
    {
      fprintf(stderr, "matrix_copy: invalid args\n");
    }
  for (i = 0; i < SRC->rows*SRC->cols; i++)
    DEST->el[i] = SRC->el[i];
}


/* matrix_duplicate ---
 *    Creates a duplicate of the matrix A
 */
Matrix *
matrix_duplicate(const Matrix *A)
{
  Matrix *newmat;

  newmat = matrix_create(A->rows, A->cols);
  matrix_copy(newmat, A);
  return(newmat);
}


/* matrix_print ---
 *    Prints out a matrix A.
 */
void
matrix_print(const Matrix *A, const char *name)
{
  int i,j;

  if (name) printf("%s(r%d,c%d) =\n\n", name, A->rows, A->cols);
  for (i = 0; i < A->rows; i++)
    {
      printf("\t");
      for (j = 0; j < A->cols; j++)
	{
	  if (ISZERO(MIDX(A,i,j)))
	    printf("        0 ");
	  else
	    printf("%9.4f ", MIDX(A, i, j));
	}
      printf("\n");
    }
  printf("\n");  
}

/* matrix_slu ---
 *   Factors A = LU.  Takes A and returns the L and U factors. 
 *   L and U must exist and must be the same size as A.
 *   A, L, and U must all be square.
 *   Returns 0 on failure, else 1.
 *
 *   Algorithm from Page 83 of Strang, "Introduction to Linear Algebra"
 */
int
matrix_slu(Matrix *L, Matrix *U, const Matrix *Aarg)
{
  int n, k, i, j;
  Matrix *A;

  if (!Aarg || !L || !U)
    { fprintf(stderr,"matrix_slu: passed null arg\n");
      return(0); }
  A = matrix_duplicate(Aarg);
  n = A->cols;
  if ((n != A->rows) 
      || (n != L->rows) || (n != L->cols) 
      || (n != U->rows) || (n != U->cols))
    { fprintf(stderr,
	      "matrix_slu: matrices not square or dimensions do not agree.\n");
      return(0); }

  for (k = 0; k < n; k++)
    {
      if (ISZERO(MIDX(A,k,k)))
	{
	  fprintf(stderr, "matrix_slu: small pivot in column %d\n", k);
	  return(0);
	}
      MIDX(L,k,k) = 1;

      for (i = k+1; i < n; i++)
	{
	  MIDX(L,i,k) = MIDX(A,i,k) / MIDX(A,k,k);
	  for (j = k+1; j < n; j++)
	    {
	      MIDX(A,i,j) = MIDX(A,i,j) - MIDX(L,i,k) * MIDX(A,k,j);	      
	    }
	}
      for (j = k; j < n; j++)
	{
	  MIDX(U,k,j) = MIDX(A,k,j);
	}
    }
  matrix_destroy(A);
  return(1);
}


/* matrix_slv ---
 *   Solve for x in Ax = b using L and U from slu(A).
 *   A, b, and x must exist.  A must be NxN and b and x must be Nx1.
 *   Returns 0 on failure, else 1.
 *
 *   Algorithm from Page 83 of Strang, "Introduction to Linear Algebra"
 */
int
matrix_slv(Matrix *x, const Matrix *A, const Matrix *b)
{
  int n, k, j;
  Matrix *L, *U, *c;
  double s, t;

  if (!A || !x || !b)
    { fprintf(stderr,"matrix_slv: passed null arg\n");
      return(0); }
  n = A->cols;
  if ((n != A->rows) 
      || (n != b->rows) || (1 != b->cols) 
      || (n != x->rows) || (1 != x->cols))
    { fprintf(stderr, "matrix_slv: invalid matrix size.\n");
      return(0); }
  L = matrix_create(n,n);
  U = matrix_create(n,n);

  if (0 == matrix_slu(L,U,A))
    { fprintf(stderr, "matrix_slv: matrix_slu failed.\n");
      matrix_destroy(L); matrix_destroy(U);
      return(0); }

  c = matrix_create(n,1);

  /* Forward elimination to solve Lc = b.
   * L is the lower triangular with 1's on the diagonal */
  for (k = 0; k < n; k++)
    {
      s = 0.0;
      for (j = 0; j <= k-1; j++)
	{
	  s += MIDX(L,k,j)*MIDX(c,j,0);
	}
      MIDX(c,k,0) = MIDX(b,k,0) - s;
    }

  /* Back substitution to solve Ux = c.
   * U is upper triangular with nonzeros on the diagonal. */
  for (k = n-1; k >= 0 ; k--)
    {
      t = 0;
      for (j = k+1; j < n; j++)
	{
	  t += MIDX(U,k,j)*MIDX(x,j,0);
	}
      MIDX(x,k,0) = (MIDX(c,k,0) - t) / MIDX(U,k,k);
    }

  matrix_destroy(c); 
  matrix_destroy(L); 
  matrix_destroy(U);
  return(1);
}


/* matrix_mult ---
 *   Performs C = A*B.  
 *   A, B, and C must exist and must be of the proper size.
 *   Returns 0 on failure, else 1.
 */
int
matrix_mult(Matrix *C, const Matrix *A, const Matrix *B)
{
  int i, j, n;
  double accum;

  if (!A || !B || !C)
    { fprintf(stderr,"matrix_mult: passed null arg\n");
      return(0); }

  if ((A->cols != B->rows) 
      || (A->rows != C->rows) 
      || (B->cols != C->cols))
    { fprintf(stderr, "matrix_mult: invalid matrix size.\n");
      return(0); }

  for (i = 0; i < C->rows; i++)
    for (j = 0; j < C->cols; j++)
      {
	accum = 0.0;
	for (n = 0; n < A->cols; n++)
	  accum += MIDX(A,i,n) * MIDX(B,n,j);
	MIDX(C,i,j) = accum;
      }
  return(1);
}


/* matrix_transpose ---
 *   Transposes the matrix A and puts the result into B.
 *   A and B must exist and must be of the proper size.
 *   Returns 0 on failure, else 1.
 */
int
matrix_transpose(Matrix *B, const Matrix *A)
{
  int i, j;

  if (!A || !B)
    { fprintf(stderr,"matrix_transpose: passed null arg\n");
      return(0); }

  if ((A->rows != B->cols) 
      || (A->cols != B->rows))
    { fprintf(stderr, "matrix_transpose: invalid matrix size.\n");
      return(0); }

  for (i = 0; i < B->rows; i++)
    for (j = 0; j < B->cols; j++)
      {
	MIDX(B,i,j) = MIDX(A,j,i);
      }
  return(1);
}


/* matrix_lsq ---
 *   Solve for x in Ax = b using least squares.
 *   Actually solves A'Ax = A'b.
 *   A, b, and x must exist.  A must be MxN and b must be Mx1 
 *   and x must be Nx1.
 *   Returns 0 on failure, else 1.
 */
int
matrix_lsq(Matrix *x, const Matrix *A, const Matrix *b)
{
  Matrix *Atrans, *AA, *Ab;
  int retval = 1;

  if (!A || !x || !b)
    { fprintf(stderr,"matrix_lsq: passed null arg\n");
      return(0); }
  if ((A->rows != b->rows) || (A->cols != x->rows) 
      || (1 != b->cols) || (1 != x->cols))
    { fprintf(stderr, "matrix_lsq: invalid matrix size.\n");
      return(0); }
  
  Atrans = matrix_create(A->cols, A->rows);
  AA = matrix_create(A->cols, A->cols);
  Ab = matrix_create(A->cols, 1);

  matrix_transpose(Atrans, A);
  matrix_mult(AA, Atrans, A);

  matrix_mult(Ab, Atrans, b);

  if (0 == matrix_slv(x, AA, Ab))
    { fprintf(stderr, "matrix_lsq: matrix_slv failed.\n");
      retval = 0; }

  matrix_destroy(Atrans); 
  matrix_destroy(AA); 
  matrix_destroy(Ab);
  return(retval);
}
