/* $Header: /home/nygren/courses/6.866/project/src/RCS/opflow.c,v 1.2 1996/02/17 19:36:41 nygren Exp $ */
/*
 * FILE:    opflow.c
 * AUTHOR:  Erik Nygren.
 * PURPOSE: Routines for generating the optical flow.
 *          Written for 6.866 term project, Fall 1995.
 */

#include <stdio.h>
#include "image.h"
#include "opflow.h"

DerivField *
derivfield_make(int x, int y)
{
  DerivField *new;
  new = (DerivField*)xmalloc("derivfield_make",sizeof(DerivField));
  new->xres = x;
  new->yres = y;
  new->data = (double*)xcalloc("derivfield_make", x*y, sizeof(double));
  return(new);
}

/* Macro for getting a double representing brightness. */
#define EV(x,y,ip) (double)(iplane_getval((ip),(x),(y)))

/* Makes the Ex, Ey, and Et derivative fields between the two image planes.
 * Algorithm from pg288 in Horn's Robot Vision.
 * Requires that ip0 and ip1 are the same resolution.
 * The resulting plane is smaller by one pixel in both x and y dirs.
 */
DerivField *
derivfield_make_Ex(IPlane *ip0, IPlane *ip1)
{
  DerivField *df;
  double *dfcur;
  int x, y;
  const double delx=1.0;

  df = derivfield_make(ip0->xres-1, ip0->yres-1);
  dfcur = df->data;
  for (y = 0; y < df->yres; y++)
    for (x = 0; x < df->xres; x++)
      {
	*dfcur =
	  ((EV(x+1,y,ip0) + EV(x+1,y,ip1) + EV(x+1,y+1,ip0) + EV(x+1,y+1,ip1))
	   - 
	   (EV(x,y,ip0) + EV(x,y,ip1) + EV(x,y+1,ip0) + EV(x,y+1,ip1)))
	    /(4.0*delx);
	dfcur++;
      }
  return(df);
}

DerivField *
derivfield_make_Ey(IPlane *ip0, IPlane *ip1)
{
  DerivField *df;
  double *dfcur;
  int x, y;
  const double dely=1.0;

  df = derivfield_make(ip0->xres-1, ip0->yres-1);
  dfcur = df->data;
  for (y = 0; y < df->yres; y++)
    for (x = 0; x < df->xres; x++)
      {
	*dfcur =
	  ((EV(x,y+1,ip0) + EV(x,y+1,ip1) + EV(x+1,y+1,ip0) + EV(x+1,y+1,ip1))
	   - 
	   (EV(x,y,ip0) + EV(x,y,ip1) + EV(x+1,y,ip0) + EV(x+1,y,ip1)))
	    /(4.0*dely);
	dfcur++;
      }
  return(df);
}

DerivField *
derivfield_make_Et(IPlane *ip0, IPlane *ip1)
{
  DerivField *df;
  double *dfcur;
  int x, y;
  const double delt=1.0;

  df = derivfield_make(ip0->xres-1, ip0->yres-1);
  dfcur = df->data;
  for (y = 0; y < df->yres; y++)
    for (x = 0; x < df->xres; x++)
      {
	*dfcur =
	  ((EV(x,y,ip1) + EV(x,y+1,ip1) + EV(x+1,y,ip1) + EV(x+1,y+1,ip1))
	   - 
	   (EV(x,y,ip0) + EV(x,y+1,ip0) + EV(x+1,y,ip0) + EV(x+1,y+1,ip0)))
	    /(4.0*delt);
	dfcur++;
      }
  return(df);
}

double
derivfield_getval(DerivField *df, int x, int y)
{
  if (x >= 0  &&  y >= 0  &&  x < df->xres  &&  y < df->yres)
    return((df->data)[x+y*df->xres]);
  else
    {
      fprintf(stderr, "derivfield_getval: %d,%d out of bounds\n",x,y);
      return(0);
    }
}

/* Makes a new optical flow field */
OpFlow *
opflow_make(int x, int y)
{
  OpFlow *new;
  int i;

  new = (OpFlow*)xmalloc("opflow_make",sizeof(OpFlow));
  new->xres = x;
  new->yres = y;
  new->u = (double*)xcalloc("opflow_make", x*y, sizeof(double));
  new->v = (double*)xcalloc("opflow_make", x*y, sizeof(double));
  for (i=0; i<x*y; i++)
    {
      (new->u)[i] = 0.0;
      (new->v)[i] = 0.0;
    }
  return(new);
}

void 
opflow_destroy(OpFlow *of)
{
  if (of && of->u && of->v)
    {
      xfree(of->u);
      xfree(of->v);
      xfree(of);
    }
}

double 
opflow_getu(OpFlow *of, int x, int y)
{
  if (x >= 0  &&  y >= 0  &&  x < of->xres  &&  y < of->yres)
    return((of->u)[x+y*of->xres]);
  else
    {
      fprintf(stderr, "opflow_getu: %d,%d out of bounds\n",x,y);
      return(0);
    }
}

double 
opflow_getv(OpFlow *of, int x, int y)
{
  if (x >= 0  &&  y >= 0  &&  x < of->xres  &&  y < of->yres)
    return((of->v)[x+y*of->xres]);
  else
    {
      fprintf(stderr, "opflow_getv: %d,%d out of bounds\n",x,y);
      return(0);
    }
}


/* Generates a field of the average surrounding values.
 * Passed a field and allocs and returns a field of the averages
 * Average comes from (u(k,l)+u(k+1,l)+u(k,l+1)+u(k+1,l+1))/4
 */
OpFlow *
opflow_gen_average_field(OpFlow *of)
{
  int x,y;
  int xr, yr;  /* x and y res */
  OpFlow *av;  /* Average */

  xr = of->xres;
  yr = of->yres;
  printf("opflow_gen_average_field:  xr=%d, yr=%d\n",xr,yr);

  av = opflow_make(xr,yr);

  for (y = 0; y < yr-1; y++)
    for (x = 0; x < xr-1; x++)
      {
	(av->u)[x+xr*y] = (  (of->u)[x   + xr*(y)]
			   + (of->u)[x+1 + xr*(y)]
			   + (of->u)[x   + xr*(y+1)]
			   + (of->u)[x+1 + xr*(y+1)]
			   )/4.0;
	(av->v)[x+xr*y] = (  (of->v)[x   + xr*(y)]
			   + (of->v)[x+1 + xr*(y)]
			   + (of->v)[x   + xr*(y+1)]
			   + (of->v)[x+1 + xr*(y+1)]
			   )/4.0;
      }
  /* Boundaries - do the best we can */
  for (x = 0, y = yr-1; x < xr-1; x++)
    {
      (av->u)[x+xr*y] = (  (of->u)[x   + xr*y]
			 + (of->u)[x+1 + xr*y]
			 )/2.0;     
      (av->v)[x+xr*y] = (  (of->v)[x   + xr*y]
			 + (of->v)[x+1 + xr*y]
			 )/2.0;     
    }
  for (y = 0, x = xr-1; y < yr-1; y++)
    {
      (av->u)[x+xr*y] = (  (of->u)[x   + xr*y]
			 + (of->u)[x+1 + xr*y]
			 )/2.0;     
      (av->v)[x+xr*y] = (  (of->v)[x   + xr*y]
			 + (of->v)[x+1 + xr*y]
			 )/2.0;     
    }
  (av->v)[xr-1 + xr*(yr-1)] = (of->v)[xr-1 + xr*(yr-1)];

  return(av);
}


/* Does an iteration of optical flow.
 * Algorithm from pg288 in Horn's Robot Vision.
 * Allocs and returns a new optical flow field.
 */
OpFlow *
opflow_doiter(OpFlow *of0, DerivField *ExField, DerivField *EyField,
	      DerivField *EtField, double Lambda)
{
  OpFlow *of1, *ofav;
  int x, y, xr, yr;
  double Ex, Ey, Et, uav, vav;

  xr = of0->xres;
  yr = of0->yres;

  of1 = opflow_make(xr, yr);
  ofav = opflow_gen_average_field(of0);
  
  for (y = 0; y < yr; y++)
    for (x = 0; x < xr; x++)
      {
	uav = opflow_getu(ofav, x, y);
	vav = opflow_getu(ofav, x, y);
	Ex = derivfield_getval(ExField,x,y);	
	Ey = derivfield_getval(EyField,x,y);	
	Et = derivfield_getval(EtField,x,y);	

	(of1->u)[x+xr*y] 
	  = uav
	    - Ex * Lambda 
	      * (Ex*uav + Ey*vav + Et) / (1 + Lambda*(Ex*Ex + Ey*Ey));
	(of1->v)[x+xr*y] 
	  = vav
	    - Ey * Lambda 
	      * (Ex*uav + Ey*vav + Et) / (1 + Lambda*(Ex*Ex + Ey*Ey));	
      }

  opflow_destroy(ofav);
  return(of1);
}

