/* $Header: /home/nygren/courses/6.866/project/src/RCS/projxform.c,v 1.5 1996/03/03 02:07:22 nygren Exp $ */
/*
 * FILE:    projxform.c
 * AUTHOR:  Erik Nygren.
 * PURPOSE: Routines for transforming images.
 *          Written for 6.866 term project, Fall 1995.
 */

#include <stdio.h>
#include <stdlib.h>
#include "image.h"
#include "projxform.h"
#include "linalg.h"

ProjXForm *
projxform_create(double D, double E, double F, double K, double L, double M,
		 double Gx, double Hx, double Gy, double Hy)
{
  ProjXForm *pxf;

  pxf = (ProjXForm*)xcalloc("projxform_create", 1, sizeof(ProjXForm));
  pxf->D = D;     pxf->E = E;     pxf->F = F;
  pxf->K = K;     pxf->L = L;     pxf->M = M;
  pxf->Gx = Gx;     pxf->Hx = Hx;
  pxf->Gy = Gy;     pxf->Hy = Hy;
  return(pxf);
}

/* Creates an identity transform (bx = ax, by = ay) */
ProjXForm *
projxform_create_identity(void)
{
  ProjXForm *pxf;

  pxf = (ProjXForm*)xcalloc("projxform_create_identity", 1, sizeof(ProjXForm));
  pxf->D = 1;     pxf->E = 0;     pxf->F = 0;
  pxf->K = 0;     pxf->L = 1;     pxf->M = 0;
  pxf->Gx = 0;    pxf->Hx = 0;
  pxf->Gy = 0;    pxf->Hy = 0;
  return(pxf);
}

/* projxform_compose ---
 *    Composes two projective transforms.  If passed a transform
 *    from A to B and a transform from B to C, creates and returns
 *    a transform from A to C.  */
ProjXForm *
projxform_compose(ProjXForm *AB, ProjXForm *BC)
{
  ProjXForm *AC;		/* The transform to return. */
  double     denom;		/* Divides every term by this to get
				 * into standard form. */

  double D1, E1, F1, G1, H1, K1, L1, M1;
  double D2, E2, F2, G2, H2, K2, L2, M2;
  double D3, E3, F3, G3, H3, K3, L3, M3;

  /* Just copying things over to make things cleaner */
  D1 = AB->D; E1 = AB->E; F1 = AB->F; K1 = AB->K; L1 = AB->L; M1 = AB->M;
  G1 = (AB->Gx + AB->Gy)/2.0;   H1 = (AB->Hx + AB->Hy)/2.0;
  D2 = BC->D; E2 = BC->E; F2 = BC->F; K2 = BC->K; L2 = BC->L; M2 = BC->M;
  G2 = (BC->Gx + BC->Gy)/2.0;   H2 = (BC->Hx + BC->Hy)/2.0;

  denom = G2*F1 + H2*M1 + 1;

  D3 = (D2*D1 + E2*K1 + F2*G1) / denom;
  E3 = (D2*E1 + E2*L1 + F2*H1) / denom;
  F3 = (D2*F1 + E2*M1 + F2) / denom;

  K3 = (K2*D1 + L2*K1 + M2*G1) / denom;
  L3 = (K2*E1 + L2*L1 + M2*H1) / denom;
  M3 = (K2*F1 + L2*M1 + M2) / denom;
  
  G3 = (G2*D1 + H2*K1 + G1) / denom;
  H3 = (G2*E1 + H2*L1 + H1) / denom;
  
  AC = projxform_create(D3, E3, F3, K3, L3, M3, G3, H3, G3, H2);

  return(AC);
}

/* projxform_xform ---
 *    Pushes points through the projective transform.
 *    Returns 1 if success, returns 0 if the results 
 *    are invalid (due to the points being behind the COP.
 */
INLINE int
projxform_xform(ProjXForm *pxf,
		double ax, double ay, double *bx, double *by)
{
  double denomx, denomy;

  denomx = (ax*pxf->Gx + ay*pxf->Hx + 1);
  denomy = (ax*pxf->Gy + ay*pxf->Hy + 1);
  if (denomx > 0 && denomy > 0)
    {
      *bx = (ax*pxf->D + ay*pxf->E + pxf->F) / denomx;
      *by = (ax*pxf->K + ay*pxf->L + pxf->M) / denomy;
      return(1);
    }
  return(0);
}

/* projxform_destroy ---
 *    Destroys the projective transform object pxf 
 */
extern void
projxform_destroy(ProjXForm *pxf)
{
  xfree(pxf);
}



/* Does a projective transform on an image.  Passed the image "B"
 * (which has the points (bx,by) and the image to transform onto
 * (ax,ay) */
void
projxform_xform_image(ProjXForm *pxf, Image *imA, Image *imB)
{
  int axmin, aymin, axmax, aymax;
  int ax, ay;
  double bx, by;
  int xabs, yabs;
  double blurval;		/* How much to get from the B image */
  ulong32 val, Aval;
  int rendered_on_scanline;  /* Have we already rendered on this scanline? */

  axmin = -imA->xorigin;
  aymin = -imA->yorigin;
  axmax = imA->xres - imA->xorigin;
  aymax = imA->yres - imA->yorigin;
  
  for (ay = aymin; ay<aymax; ay++)
    {
      rendered_on_scanline = 0;
      for (ax = axmin; ax<axmax; ax++)
	{
	  /* Transform the points.  If invalid, move on. */
	  if (!projxform_xform(pxf, ax, ay, &bx, &by))
	    continue;
	  val = image_getval_rel(imB, bx, by);
	  if (val != -1)
	    {
	      rendered_on_scanline = 1;
#define DOBLUR
#define BLURWIDTH 20
#define DBLURWIDTH ((double)BLURWIDTH)
#ifdef DOBLUR
	      xabs = imB->xorigin + (int)bx;
	      yabs = imB->yorigin + (int)by;
	      blurval = 1.0;
	      if (xabs < BLURWIDTH)
		blurval = MIN(blurval, 
			      (double)(xabs)/DBLURWIDTH);
	      if (yabs < BLURWIDTH)
		blurval = MIN(blurval, (double)(yabs)/DBLURWIDTH);
	      if (imB->xres-xabs < BLURWIDTH)
		blurval = MIN(blurval, (double)(imB->xres-xabs)/DBLURWIDTH);
	      if (imB->yres-yabs < BLURWIDTH)
		blurval = MIN(blurval, (double)(imB->yres-yabs)/DBLURWIDTH);
	      if (blurval < 1.0)
		{
		  /* NOTE: This only works with grey images */
		  Aval = image_getval_rel(imA, ax, ay);
		  val = 0xff&(ulong32)(MIN(255.0,
					   (double)(0xff&val)*blurval +
					   (double)(0xff&Aval)*(1.0-blurval)));
		  /*printf("%f,%f blurval %f\n",bx,by,blurval);*/
		}
#endif
	      image_setval_rel(imA, ax, ay, val);
	    }
	  else if (rendered_on_scanline)
	    {
	      break;		/* Skip to next scanline */
	    }
	}
    }
}

/* Does a projective transform on an image.  Passed the image "B"
 * (which has the points (bx,by) and the image to transform onto
 * (ax,ay).  This version should be better optimized but does no blurring. */
void
projxform_xform_image_quick(ProjXForm *pxf, Image *imA, Image *imB)
{
  int Axres, Ayres, Bxres, Byres;
  int Axoff, Ayoff, Bxoff, Byoff;
  int ax, ay;
  double bx, by;
  int ibx, iby;
  int rendered_on_scanline;  /* Have we already rendered on this scanline? */
  uchar *Adata, *Bdata;

  if (imA->np!=1 || imB->np!=1)
    { fprintf(stderr,
	      "projxform_xform_image_quick: only grayscale supported\n");
      return; }

  Axres = imA->xres; Ayres = imA->yres; 
  Bxres = imB->xres; Byres = imB->yres;
  Axoff = imA->xorigin; Ayoff = imA->yorigin; 
  Bxoff = imB->xorigin; Byoff = imB->yorigin;
  Adata = imA->p[0]->data;
  Bdata = imB->p[0]->data;

  for (ay = 0; ay<Ayres; ay++)
    {
      rendered_on_scanline = 0;
      for (ax = 0; ax<Axres; ax++)
	{
	  /* Transform the points.  If invalid, move on. */
	  if (!projxform_xform(pxf, ax-Axoff, ay-Ayoff, &bx, &by))
	    continue;
	  ibx = ((int)bx) + Bxoff;  
	  iby = ((int)by) + Byoff;  
	  if (ibx >= 0 && iby >= 0 && ibx < Bxres && iby < Byres)
	    {
	      rendered_on_scanline=1;
	      Adata[ax+ay*Axres] = Bdata[ibx+iby*Bxres];
	    }
	  else if (rendered_on_scanline)
	    {
	      break;		/* Skip to next scanline */
	    }
	}
    }
}


/* Creates a PointPairList, a list of point pairs. */
PointPairList *
pointpairlist_create(void)
{
  PointPairList *ppl;

  ppl = (PointPairList*)xcalloc("pointpairlist_create", 1,
				sizeof(PointPairList));
  ppl->head = NULL;
  ppl->numpairs = 0;
  return(ppl);
}

/* Adds a pair of points to a PointPairList */
void
pointpairlist_addpair(PointPairList *ppl, double ax, double ay,
		      double bx, double by)
{
  PointPair *newpair;

  newpair = (PointPair*)xcalloc("pointpairlist_addpair", 1,
				sizeof(PointPair));
  newpair->ax = ax;
  newpair->ay = ay;
  newpair->bx = bx;
  newpair->by = by;
  /* Now link the pair into the list */
  newpair->next = ppl->head;
  ppl->head = newpair;
  ppl->numpairs++;
}

/* Flips a and b in the point pairs and returns a new pointpairlist. */
PointPairList *
pointpairlist_flipab(PointPairList *pplin)
{
  PointPairList *pplout;
  PointPair     *pCur;

  pplout = pointpairlist_create();
  for (pCur = pplin->head; pCur; pCur = pCur->next)
    {
      pointpairlist_addpair(pplout, pCur->bx, pCur->by, pCur->ax, pCur->ay);
    }
  return(pplout);
}


/* pointpairlist_genprojxform ---
 *    Calculates a projective transform by applying least squares
 *    to the point pair list.  Returns the projective transform
 *    or NULL if unable to.
 *
 * Uses:
 *
 *  bx = [ax ay 1 -ax*bx -bx*ay] * [D E F G H]'
 *
 *  by = [ax ay 1 -ax*by -by*ay] * [K L M G H]'
 *
 * Which is:
 *
 *  BX = AX * XParams
 *  BY = AY * YParams
 *
 */
ProjXForm *
pointpairlist_genprojxform(PointPairList *ppl)
{
  double D, E, F, K, L, M, GX, HX, GY, HY;
  Matrix *BX, *BY, *AX, *AY, *XParams, *YParams;
  int np, i, retval=1;
  PointPair *pCur;
  ProjXForm *pxf;

  np = ppl->numpairs;
  if (np < 5)
    { fprintf(stderr, "pointpairlist_genprojxform: not enough points\n");
      return(NULL); }
  
  BX = matrix_create(np, 1);
  BY = matrix_create(np, 1);
  XParams = matrix_create(5, 1);
  YParams = matrix_create(5, 1);
  AX = matrix_create(np, 5);
  AY = matrix_create(np, 5);
  
  for (i=0, pCur=ppl->head; pCur; i++, pCur=pCur->next)
    {
      MIDX(BX,i,0) = pCur->bx;
      MIDX(BY,i,0) = pCur->by;

      MIDX(AX,i,0) = pCur->ax;
      MIDX(AX,i,1) = pCur->ay;
      MIDX(AX,i,2) = 1.0;
      MIDX(AX,i,3) = -1.0 * pCur->ax * pCur->bx;
      MIDX(AX,i,4) = -1.0 * pCur->bx * pCur->ay;

      MIDX(AY,i,0) = pCur->ax;
      MIDX(AY,i,1) = pCur->ay;
      MIDX(AY,i,2) = 1.0;
      MIDX(AY,i,3) = -1.0 * pCur->ax * pCur->by;
      MIDX(AY,i,4) = -1.0 * pCur->by * pCur->ay;
    }

#ifdef DEBUG
  matrix_print(XParams, "XParams");
  matrix_print(AX, "AX");
  matrix_print(BX, "BX");
#endif /* DEBUG */

  if (!matrix_lsq(XParams, AX, BX))
    { fprintf(stderr,
	      "pointpairlist_genprojxform: matrix_lsq failed on xparams\n");
      retval = 0; }
  if (!matrix_lsq(YParams, AY, BY))
    { fprintf(stderr,
	      "pointpairlist_genprojxform: matrix_lsq failed on yparams\n");
      retval = 0; }
  
  D = MIDX(XParams, 0, 0);
  E = MIDX(XParams, 1, 0);
  F = MIDX(XParams, 2, 0);
  GX = MIDX(XParams, 3, 0);
  HX = MIDX(XParams, 4, 0);
  K = MIDX(YParams, 0, 0);
  L = MIDX(YParams, 1, 0);
  M = MIDX(YParams, 2, 0);
  GY = MIDX(YParams, 3, 0);
  HY = MIDX(YParams, 4, 0);

  pxf = projxform_create(D,E,F,K,L,M,GX,HX,GY,HY);

#ifdef DEBUG
  printf("D %f  E %f  F %f\n", D, E, F);
  printf("K %f  L %f  M %f\n", K, L, M);
  printf("GX %f  GY %f\n", GX, GY);
  printf("HX %f  HY %f\n", HX, HY);
#endif /* DEBUG */

  matrix_destroy(BX);       matrix_destroy(BY);
  matrix_destroy(AX);       matrix_destroy(AY);
  matrix_destroy(XParams);  matrix_destroy(YParams);

  if (retval)
    return(pxf);
  else
    return(NULL);
}
