/*************************************************************************\
 * FILE:    vlm_vidwin.c                                                 *
 * AUTHOR:  Erik Nygren <nygren@mit.edu>                                 *
 * CREATED: 29-Jun-93                                                    *
 * PURPOSE: Puts video into window                                       *
\*************************************************************************/ 

#include <stdio.h>               
#include <X11/Xlib.h>            /* Xlib include files */
#include <X11/Xutil.h>           /* Must have these included */
#include <X11/Xatom.h>
#include "vlmotion.h"

static XVisualInfo visinfo;
Display *display;       /* The display */
int      Xscreen;       /* The screen  */
Window   vidwin;        /* Window for displaying video */
Window   vidwinparent;  /* The parent to vidwin */
GC       gfxcontext;    /* Graphics context */
XImage  *vwximage;      /* VidWin's XImage structre */

/****============= GET VISUAL =============****/

Visual *get_visual(dpy)
     Display *dpy;
{
  int screen=0;

  screen=DefaultScreen(dpy);
  return(DefaultVisual(dpy, screen));
}

/****============= SURE MALLOC ============****/

char *sure_malloc(size, msg)
     unsigned size;
     char *msg;
{
  char *pt;

  if((pt = (char *)malloc(size)) == NULL)
    {
      fprintf(stderr, "%s\n", msg);
      CleanExit(1);
    }
  else
    return(pt);
}

/****============= OPEN VIDWIN ============****/

int OpenVidWin(thedisplay, parentwin, xoff, yoff, width, height,
	       vlm_vidmode)
     Display *thedisplay;
     Window parentwin;     /* The parent to this window          */
     int xoff, yoff;        /* X and Y offsets in parent window   */
     int width, height;     /* The width and height of the vidwin */
     int vlm_vidmode;  	    /* Either VLM_GRAY or VLM_MULTICHANNEL or VLM_COLOR */
{
  int screen;
  unsigned long black_color, white_color;
  static Colormap colmap = 0;
  Visual *visual;
  int depth;
  static XSetWindowAttributes xswa;
  Screen *SCREEN;
  register unsigned int k;
  static XColor colors[256];
  unsigned long color_planes[1],color_pixels[256];
  Atom ATOM_WM_COLMAP_WIN;

  
  display = thedisplay;
  vidwinparent = parentwin;
  xsize = width;
  ysize = height;

  screen = DefaultScreen(display);

  /**** SET UP COLORS ****/
  black_color = BlackPixel(display, screen);
  white_color = WhitePixel(display, screen);

  depth = DefaultDepth(display,screen); 
#ifdef DEBUG
  printf("Display has depth %d\n", depth);
#endif
  if(depth < 8)
    {
      fprintf(stderr,
	      "Sorry, %s is a %d-bit depth screen, only 8/24-bit depth is required\n", 
	      display, depth);
      CleanExit(-1);
    }
#ifdef USETRUECOLOR
  /* May not work */
  else if (XMatchVisualInfo(display,screen,8,TrueColor,&visinfo))   /* Check for TrueColor */
    {
      colmap = NULL;
      depth = 8;
      printf("Using 8-bit TrueColor\n");
    }
#endif

  /* Get a 256 color PseudoColor visual */
  else if (XMatchVisualInfo(display,screen,8,PseudoColor,&visinfo))
    {
      colmap = XCreateColormap(display, parentwin,
			       visinfo.visual, 
			       AllocAll);
      /* Set up colormap for color.  Bits are BBGGGRRR */
      if (vlm_vidmode == VLM_COLOR)
	{
	  for(k=0; k<256; k++)
	    {
	      colors[k].pixel = k;
	      colors[k].red = 	255*(((0x07 & (unsigned char)k) << 5)
		| ((0x07 & (unsigned char)k) << 2));
	      colors[k].green =	255*(((0x38 & (unsigned char)k) << 2)
		| ((0x38 & (unsigned char)k) >> 1));
	      colors[k].blue =	255*(((0xC0 & (unsigned char)k))
		| ((0xC0 & (unsigned char)k)>>2)
		  | ((0xC0 & (unsigned char)k)>>4)
		    | ((0xC0 & (unsigned char)k)>>6));
	      colors[k].flags = DoRed | DoGreen | DoBlue;
	    }	  
	}
      /* Setup colormap for grayscale */
      else
	{
	  for(k=0; k<256; k++)
	    {
	      colors[k].pixel = k;
	      colors[k].red = colors[k].green = colors[k].blue = (u_short)(k*65000/256);
	      colors[k].flags = DoRed | DoGreen | DoBlue;
	    }
	}
      XStoreColors(display,colmap,colors,256);

      /* Get window attributes */
      xswa.colormap = colmap;
#ifdef DEBUG
      printf("Using 8-bit PseudoColor\n");
#endif
    }	
  else
    {
      fprintf(stderr,
	      "Sorry, apropriate visuals are not available on this display...\n");
      CleanExit(-1);
    }

  vidwin = XCreateWindow (display, parentwin,
			  xoff, yoff, xsize, ysize, 1, depth, InputOutput,
			  visinfo.visual, (colmap ? CWColormap : 0), &xswa );
  XSetWindowBackground(display, vidwin, white_color);

  if (colmap)  /* If not TrueColor */
    {
      /* Set colormap for this window and for parent window */
      XSetWindowColormap(display, vidwin, colmap);
      XSetWindowColormap(display, parentwin, colmap);

      /* Tell windowmanager that subwindow has its own colormap */
      ATOM_WM_COLMAP_WIN = XInternAtom(display,"WM_COLMAP_WINDOWS", False);
      XChangeProperty(display,parentwin,ATOM_WM_COLMAP_WIN,XA_WINDOW,32,
		      PropModeReplace,
		      (unsigned char *)&vidwin, 1);
      XSync(display, True);
    }

  XMapWindow(display, vidwin); 
  XFlush(display);

  gfxcontext = XCreateGC (display, vidwin, 0, NULL);
  XSetBackground(display, gfxcontext, white_color);
  XSetForeground(display, gfxcontext, black_color);
#ifdef DEBUG
  printf("Opened and mapped window for video...\n");
#endif
  vwximage = XCreateImage(display, visinfo.visual, 8, ZPixmap, 0, NULL,
			width, height, 8, 0);
#ifdef OMIT
  vwximage->bitmap_unit = 8;
  vwximage->bits_per_pixel = 8;
  vwximage->red_mask = 0xffff;
  vwximage->green_mask = 0xffff;
  vwximage->blue_mask = 0xffff;
#endif

}

/****============= UPDATE VIDWIN ============****/

void UpdateVidWin(int width, int height, char *imagedata)
{
  vwximage->data = imagedata;
  XPutImage(display, vidwin, gfxcontext, vwximage, 0, 0, 0, 0, width, height);
  XFlush(display);
}

/****============= CLOSE VIDWIN ============****/

void CloseVidWin()
{
  vwximage->data = NULL;
  XDestroyImage(vwximage);
}






