/*************************************************************************\
 * FILE:    vlmotion.c                                                   *
 * AUTHOR:  Erik Nygren <nygren@mit.edu>                                 *
 * CREATED: 29-Jun-93                                                    *
 * PURPOSE: View full motion video from VideoCapture board               *
 \*************************************************************************/ 

#include <stdio.h>               
#include <X11/Xlib.h>            /* Xlib include files */
#include <X11/Xutil.h>           /* Must have these included */

#include "vlmotion.h"

#include <sgtty.h>
#include <sys/time.h>

int xoff, yoff;      /* The offset in the VideoCapture's frame of the image */
char *programname;

/**** FOR ARGUMENT PARSING ****/

int vlm_mode = VLM_COLOR;   /* Major mode */

int fork_off = FALSE;       /* Should we fork? */

int vlm_verbose_flag = FUNSET;

int vlm_sync = -1;   /* Channel to be used for sync */

char *xgeomstr;  /* Geometry string */

/**** STATISTICS STUFF ****/

/* Run a statistic through a low pass */

int loops_since_last_stat = 0;
double vlm_frame_rate;   /* Frame rate */
struct timeval last_time_val;

/**** X STUFF ****/

#define VIDBORDER 2

char *window_name = "Galileo Video";
char *icon_name = "vlmotion";
Window mainwin,   /* Main window for application */
       rootwin;   /* Screen's root window */
char *display_name = NULL;
unsigned long valuemask; 

/****========== CLEAN EXIT ============****/

void CleanExit(errno)
     int errno;
{
  XFlush(display);
  CloseVideoCapture();
  CloseVidWin();
  if (display)
    XCloseDisplay(display);
  exit(errno);
}

/****========== PRINT USAGE ===========****/

void printusage()
{
  fprintf(stderr,"Usage: %s ",programname);
  fprintf(stderr,"[-h] [-C] [-G] [-1] [-2] [-4]\n");
  fprintf(stderr,"   -C          8 Bit Dithered Color\n");
  fprintf(stderr,"   -G          8 Bit Grayscale\n");
  fprintf(stderr,"   -1          Full Size\n");
  fprintf(stderr,"   -2          Half Size\n");
  fprintf(stderr,"   -4          Quarter Size\n");
  fprintf(stderr,"\n");
}

/****========== CALCULATE TIMESTATS ==========****/

void CalculateTimeStats()
{
  struct timeval time_val;
  struct timezone time_zone;
  double delta_usec;

  if (++loops_since_last_stat > 10)
    {
      gettimeofday(&time_val,&time_zone);

      delta_usec = (double)((1000000 * (time_val.tv_sec -
					last_time_val.tv_sec)) +
			    (time_val.tv_usec - last_time_val.tv_usec));

      last_time_val.tv_sec = time_val.tv_sec;
      last_time_val.tv_usec = time_val.tv_usec;
      printf("Frame rate: %.2f Hz\n",1000000/delta_usec*loops_since_last_stat);
      loops_since_last_stat = 0;
    }
}


/****========== PARSE ARGS ==========****/

void ParseArgs(argc,argv)
     int argc;
     char *argv[];
{
  char c;

  while ((c = getopt(argc, argv, "CGhf124")) != (char)EOF) 
    {
      switch (c) 
	{  
	case 'C':
	  vlm_mode = VLM_COLOR;
	  break;
	case 'G':
	  vlm_mode = VLM_GRAY;
	  break;
	case '1':
	  zoom_denom = 1;
	  zoom_num = 1;
	  break;
	case '2':
	  zoom_denom = 2;
	  zoom_num = 1;
	  break;
	case '4':
	  zoom_denom = 4;
	  zoom_num = 1;
	  break;
	case 'f':
	  fork_off = TRUE;
	  break;
	case 'h':
	  printusage();
	  exit(0);
	  break;
	default:
	  printusage();
	  exit(0);
	  break;
	}
    }
}

/****========== PROCESS X EVENTS ==========****/

void ProcessXEvents()
{
  static XEvent report;              /* will return our events */

  if (!XEventsQueued(display,QueuedAfterReading))
    return;

  XNextEvent(display, &report);
  switch (report.type)
    {  
    case ButtonPress:
    case KeyPress:
      CleanExit(1);

      /* All events selected by StructureNotifyMask except
       * ConfigureNotify are thrown away here, since
       * nothing is done with them */
    default:
      break;
    }
}

/****========== OPEN MAIN WINDOW ============****/
/* Connect to X display and open a window */

void OpenMainWindow()
{
  static XSizeHints size_hints;
  int xpos = 100, ypos = 600;
  unsigned int dummyw, dummyh;
  int xgeomparsemask;

  if ( (display=XOpenDisplay(display_name)) == NULL ) 
    {
      fprintf( stderr, "%s: cannot connect to X server", programname);
      if (getenv("DISPLAY") == NULL)
	fprintf( stderr, ", 'DISPLAY' environment variable not set.\n");
      else
	fprintf( stderr, " %s\n", XDisplayName(display_name));
      CleanExit(-1);
    }

  Xscreen = DefaultScreen(display);

  if (xgeomstr)
    {
      xgeomparsemask = XParseGeometry(xgeomstr,&xpos,&ypos,&dummyw,&dummyh);
      if ((xgeomparsemask & XValue) || (xgeomparsemask & YValue))
	{
	  size_hints.x = xpos;
	  size_hints.y = ypos;
	  size_hints.flags = (USPosition);
	}
    }

  /* Create opaque window ... note that BlackPixel and WhitePixel
   * are macros.  */
  rootwin = RootWindow(display, Xscreen);
  mainwin = XCreateSimpleWindow (display, rootwin, xpos, ypos,
				 xsize + 2*VIDBORDER, ysize + 2*VIDBORDER, 5,
				 BlackPixel(display, Xscreen),	
				 WhitePixel(display, Xscreen));

  /* Set properties for window manager before mapping it */
  XSetStandardProperties(display, mainwin, window_name, icon_name, None,
			 0, 0, &size_hints);

  XSelectInput(display, mainwin, ExposureMask | KeyPressMask |
	       ButtonPressMask | StructureNotifyMask);
  
  /* Put window up */
  XMapWindow(display, mainwin);
}

/****========== MAIN ===========****/

main(argc,argv)
     int argc;
     char **argv;
{
  struct timezone time_zone;

  /* Process Command Line Args */
  programname = argv[0];
  ParseArgs(argc,argv);

  /* Open VideoCapture board and set for continuous monochrome grabbing */
  OpenVideoCapture(vlm_mode);

  /* Open up X display and windows */
  OpenMainWindow();
  OpenVidWin(display, mainwin, VIDBORDER, VIDBORDER, xsize, ysize,
	     vlm_mode);
  /* Prime statistics stuff */
  gettimeofday(&last_time_val,&time_zone);

  if (fork_off)
    {
      if (fork() == 0)  /* The child */
	{
	  fclose(stdin);
	  fclose(stdout);
	  fclose(stderr);
	}
      else  /* The parent */
	{
	  exit(0);
	}
    }

  /* Main event loop */
  vlMainLoop();

  CleanExit(1);
}






